// -*- C++ -*-
//
//  This file is part of CBCBSIM, the Combined Broadcast Content-Based
//  (CBCB) routing simulation.
//
//  Authors: See the file AUTHORS for full details. 
//
//  Copyright (C) 2002-2004 University of Colorado
//
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  as published by the Free Software Foundation; either version 2
//  of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307,
//  USA, or send email to serl@cs.colorado.edu.
//
// $Id: CBCBNode.cc,v 1.3 2008-04-01 17:03:04 carzanig Exp $
//
#include <cbnsim/DeliveryOracle.h>
#include <cbnsim/MatchHandler.h>
#include <cbnsim/logging.h>
#include <iostream>
#include <set>
#include <siena/ssim.h>
#include "cbcbsim/CBCBNode.h"
#include "cbcbsim/ME.h"
#include "cbcbsim/PredicateIssueTimer.h"
#include "cbcbsim/PredicateRefreshTimer.h"
#include "cbcbsim/RA.h"
#include "cbcbsim/SR.h"
#include "cbcbsim/SRSendTimer.h"
#include "cbcbsim/UR.h"

using namespace cbcbsim;
using namespace cbnsim;
using namespace ssim;
using namespace std;

#define NDEBUG(m) CBNDEBUG( "CBCBNode[" << pid() << "@" << Sim::clock() << "]: " << m )

CBCBNode :: CBCBNode( unsigned int nid, const BroadcastLayer& bl, const ProtocolOptions& options )
    : Node( nid, 0, 0 ), m_bl( bl ), m_issue_pending( false ), m_options( options ), m_route_table_dirty( false ),
    m_srur_packets( 0 )
{
}

ProcessId CBCBNode :: activate()
{
  ProcessId id = Node::activate();
  CBNASSERT( id != ssim::NULL_PROCESSID, "activate should not return a NULL processid" );
  // need to maintain the same number of 
  // emtries in the route table.
  CBNASSERT( m_route_table.empty(), "route table should be empty during activate" );
  m_route_table.push_back( Predicate() );
  m_msg_sent_counts.push_back( 0 );
  return id;
}

void CBCBNode :: add_neighbor( ssim::ProcessId id, ssim::Time length )
{
  // need to maintain the same number of 
  // emtries in the route table.
  Node::add_neighbor( id, length );
  CBNASSERT( ! m_route_table.empty(), "route table should not be empty during add_neighbor call" );
  m_route_table.push_back( Predicate() );
  m_msg_sent_counts.push_back( 0 );
}

size_t CBCBNode :: bytesize() const
{
  size_t bsize = 0;
  unsigned int size = m_route_table.size();
  for( unsigned int i = 0; i < size; ++i )
    {
      bsize += m_route_table[i].bytesize();
    }
  return bsize + m_fwd_table.bytesize();
}

void CBCBNode :: init()
{
  // predicate issue policy
  if( m_options.pip == ProtocolOptions::PIP_TIMED )
    {
      // the memory leak doesn't matter since the timer lasts until
      // the simulation is finished.
      PredicateIssueTimer* pit = new PredicateIssueTimer( *this, m_options.pip_time );
      pit->activate();
    }
  // predicate refresh policy
  if( m_options.prp == ProtocolOptions::PRP_TIMED )
    {
      PredicateRefreshTimer* prt = new PredicateRefreshTimer( *this, m_options.prp_time );
      prt->activate();
    }
  // sr send policy
  if( m_options.srsp == ProtocolOptions::SRSP_TIMED ||
      m_options.srsp == ProtocolOptions::SRSP_TIMED_COUNT )
    {
      SRSendTimer* srst = new SRSendTimer( *this, m_options.srsp_time );
      srst->activate();
    }
}

void CBCBNode :: issue_packet( unsigned int ifid, const Packet* packet, Time lag, bool count )
{
  if( count && (dynamic_cast<const SR*>( packet ) || dynamic_cast<const UR*>( packet ) ) )
    {
      m_srur_packets++;
    }
  cbnsim::Node::issue_packet( ifid, packet, lag, count );
}

void CBCBNode :: issue_predicate( bool force )
{
  NDEBUG( "****************** start issue_predicate() ******************" );
  if( m_issue_pending || force )
    {
      if( !m_issue_pending )
	{
	  NDEBUG( "predicate issue is being forced" );
	}
      propagate_ra( pid(), m_route_table[0] );
      m_issue_pending = false;
    }
  NDEBUG( "****************** end issue_predicate() ******************" );
}

void CBCBNode :: issue_sr()
{
  NDEBUG( "****************** start issue_sr() ******************" );
  // just propagate a new SR, ignoring the number of children
  propagate_sr( pid(), SR::next_number() );
  NDEBUG( "****************** end issue_sr() ******************" );
}

void CBCBNode :: process_packet( unsigned int ifid, const Packet& pkt )
{
  NDEBUG( "****************** start process_packet ******************" );
  NDEBUG( "received packet from process: " << i2p(ifid) << " (interface " << ifid << ")" );

  const RA* ra;
  const ME* me;
  const SR* sr;
  const UR* ur;
  if( (ra = dynamic_cast<const RA*>( &pkt )) )
    {
      process_ra( ifid, *ra );
    }
  else if( (me = dynamic_cast<const ME*>( &pkt )) )
    {
      process_me( ifid, *me );
    }
  else if( (sr = dynamic_cast<const SR*>( &pkt )) )
    {
      process_sr( ifid, *sr );
    }
  else if( (ur = dynamic_cast<const UR*>( &pkt )) )
    {
      process_ur( ifid, *ur );
    }
  else
    {
      CBNASSERT( false, "Unexpected event type" );
    }
  NDEBUG( "****************** end process_packet ******************" );
}

void CBCBNode :: process_me( unsigned int ifid, const ME& me )
{
  if( ! propagate_me( me.issuer(), me.message_id(), me.message() ) )
    {
      cbnsim::DeliveryOracle::log_false_delivery( me.message_id(), pid() );
    }
}

void CBCBNode :: process_ra( unsigned int ifid, const RA& ra )
{
  NDEBUG( "received RA: " << ra );
  /*
   * cp is the existing predicate for the src node
   */
  Predicate& cp( m_route_table[ifid] );
  NDEBUG( "initial predicate for process " << i2p(ifid) << ": " << cp );

  /*
   * 1) see if new predicate is already covered by
   * the existing predicate for the incoming interface.
   */
  if( siena::covers( cp, ra.predicate() ) )
    {
      NDEBUG( cp << " covers " << ra.predicate() );
      return;
    }
  
  /*
   * 2) update the route table entry for the incoming interface with
   * the new predicate.
   */
  cp.add( ra.predicate() );
  NDEBUG( "updated predicate for process " << i2p(ifid) << ": " << cp );
  NDEBUG( "marking route table dirty" );
  m_route_table_dirty = true;

  /*
   * 3) propagate the new predicate down the broadcast tree rooted
   * at the issuer's node.
   */
  propagate_ra( ra.issuer(), ra.predicate() );
}

void CBCBNode :: process_sr( unsigned int ifid, const SR& sr )
{
  // first propagate down to children
  unsigned int count = propagate_sr( sr.issuer(), sr.number() );

  // if there were no recipients to wait for, then send a UR reply to
  // parent.
  if( !count )
    {
      NDEBUG( "effective leaf node for SR: " << sr );
      
      ProcessId parent = m_bl.getParent( sr.issuer(), pid() );
      UR* ur = new UR( pid(), sr.issuer(), sr.number(), m_route_table[0] );

      NDEBUG( "responding to parent(" << parent << ") with UR: " << *ur );
      issue_packet( parent, ur );
    }
}

void CBCBNode :: process_ur( unsigned int ifid, const UR& ur )
{
  ProcessId sender = i2p(ifid);
  CBNASSERT( sender == ur.sender(), "sender calculations do not match" );

  // if we were the issuer of the SR that this is in response to
  // just set the predicate for the UR's sender
  if( ur.sr_issuer() == pid() )
    {
      NDEBUG( "Setting predicate for " << sender << " to " << ur.predicate() );

      m_route_table[ifid] = ur.predicate();
      NDEBUG( "marking route table as dirty" );
      m_route_table_dirty = true;
    }
  // otherwise we have to remove the sender from the respondees list
  // and propagate up to our parent if necessary
  else
    {
      map<unsigned int,sr_response>::iterator i = m_sr_responses.find( ur.sr_number() );
      if( i == m_sr_responses.end() )
	{
	  NDEBUG( "received unexpected UR: " << ur << ", node failed previously?" );
	  return;
	}
      
      set<ProcessId>::iterator j = (*i).second.respondees.find( sender );
      CBNASSERT( j != (*i).second.respondees.end(),
		 pid() << "@" << Sim::clock() << ": received UR from unexpected respondee: " << ur );
      
      // add this predicate to the predicate being collected
      NDEBUG( "adding predicate from " << sender << " to entry for SR number " << ur.sr_number() );
      (*i).second.pred.add( ur.predicate() );

      // erase respondee from 
      (*i).second.respondees.erase( j );

      if( (*i).second.respondees.empty() )
	{
	  NDEBUG( "all URs have been received for SR number " << ur.sr_number() );
	  // make sure to add my own predicate in there...
	  (*i).second.pred.add( m_route_table[0] );

	  // lookup parent and send
	  ProcessId parent = m_bl.getParent( ur.sr_issuer(), pid() );

	  NDEBUG( "issuing completed UR packet for SR number " << ur.sr_number() << " to " << parent );
	  issue_packet( parent, new UR( pid(), ur.sr_issuer(), ur.sr_number(), (*i).second.pred ) );

	  // erase the entry for this SR
	  m_sr_responses.erase( i );
	}
    }
}

bool CBCBNode :: propagate_me( ProcessId issuer, MessageId msg_id, const Message& msg )
{
  ME* me( NULL );
  bool matched_something = false;

  const set<ProcessId>& pids( m_bl.getChildren( issuer, pid() ) );

  // before matching update the fwd table (if its out of sync)
  if( m_route_table_dirty )
    {
      NDEBUG( "route table dirty, updating fwd table..." );
      update_fwd_table();
    }
  cbnsim::MatchHandler mh;
  m_fwd_table.match( msg, mh );

  const set<siena::if_t>& matches = mh.getIds();
  for( set<siena::if_t>::const_iterator i = matches.begin();
       i != matches.end(); ++i )
    {
      if( ! (*i) )
	{
	  // we matched ourself...
	  NDEBUG( "matched local application" );
	  matched_something = true;
	  cbnsim::DeliveryOracle::log_true_delivery( msg_id, pid() );
	}
      else
	{
	  ProcessId p( i2p( (*i) ) );
	  NDEBUG( "matched process " << p << " (interface " << (*i) << ")" );
	  /*
	   * we only propagate to neighbors that
	   * the broadcast layer said we could.
	   */
	  if( pids.find( p ) != pids.end() )
	    {
	      if( !me )
		{
		  me = new ME( pid(), issuer, msg_id, msg );
		}
	      issue_packet( (*i), me );
	      m_msg_sent_counts[(*i)]++;
	      matched_something = true;
	    }
	  else
	    {
	      NDEBUG( p << " is not child in broadcast tree rooted at " << issuer );
	    }
	}
    }
  if( !matched_something )
    {
      NDEBUG("message did not match any interfaces, dropping..." );
    }
  return matched_something;
}

void CBCBNode :: propagate_ra( ProcessId issuer, const Predicate& pred )
{
  const set<ProcessId>& children( m_bl.getChildren( issuer, pid() ) );
  NDEBUG( children.size() << " children in broadcast tree rooted at " << issuer );
  RA* ra( NULL );
  set<ProcessId>::const_iterator i;
  for( i = children.begin(); i != children.end(); ++i )
    {
      if( !ra )
	{
	  ra = new RA( pid(), issuer, pred );
	}
      NDEBUG( "propagating RA packet issued by " << issuer << " to " << (*i) );
      issue_packet( (*i), ra );
    }
}

unsigned int CBCBNode :: propagate_sr( ProcessId issuer, unsigned int number )
{
  const set<ProcessId>& children( m_bl.getChildren( issuer, pid() ) );
  NDEBUG( children.size() << " children in broadcast tree rooted at " << issuer );
  SR* sr( NULL );
  unsigned int count = 0;
  set<ProcessId>::const_iterator i;
  for( i = children.begin(); i != children.end(); ++i )
    {
      if( m_route_table[p2i(*i)].size() == 0 ||
	  (m_options.srsp == ProtocolOptions::SRSP_TIMED_COUNT &&
	   m_msg_sent_counts[p2i(*i)] < m_options.srsp_count) )
	{
	  continue;
	}
      if( !sr )
	{
	  sr = new SR( pid(), issuer, number );
	}
      NDEBUG( "propagating SR packet issued by " << issuer << " to " << (*i) );
      issue_packet( (*i), sr );
      // if we are not the issuer of this SR, but we are sending
      // down to some children we need to keep track of them.
      if( issuer != pid() )
	{
	  NDEBUG( "adding process " << (*i) << " to the respondees list of " << number );
	  m_sr_responses[number].respondees.insert( (*i) );
	}
      count++;
    }
  return count;
}

void CBCBNode :: send_message( MessageId mid, const Message& message )
{
  NDEBUG( "****************** start send_message ******************" );
  NDEBUG( "received send_message: " << message );
  propagate_me( pid(), mid, message );
  NDEBUG( "****************** end send_message   ******************" );
}

void CBCBNode :: set_predicate( const Predicate& p, unsigned int, bool )
{
  NDEBUG( "****************** start set_predicate ******************" );
  NDEBUG( "received set_predicate: " << p );
  m_route_table[0] = p;
  NDEBUG( "marking route table as dirty" );
  m_route_table_dirty = true;
  m_issue_pending = true;
  
  if( m_options.pip == ProtocolOptions::PIP_IMMEDIATE )
    {
      issue_predicate();
    }
  NDEBUG( "****************** end set_predicate ******************" );
}

void CBCBNode :: update_fwd_table()
{
  m_fwd_table.clear_recycle();

  unsigned int size = m_route_table.size();
  for( unsigned int i = 0; i < size; ++i )
    {
      NDEBUG( "ifconfig " << i << "(" << i2p( i ) << ") -> " << m_route_table[i] );
      m_fwd_table.ifconfig( i, m_route_table[i] );
    }
  m_fwd_table.consolidate();
  m_route_table_dirty = false;
}
